
/**
   File name ubitx_I2C4LT_V6_02b.ino
   This source file is under General Public License version 3.

   This version uses a built-in Si5351 library
   Most source code are meant to be understood by the compilers and the computers.
   Code that has to be hackable needs to be well understood and properly documented.
   Donald Knuth coined the term Literate Programming to indicate code that is written be
   easily read and understood.

   The xxxxxxxx is a small board that includes the Teensy 3.5 or 3.6, and using
   a 4x20 LCD display and an Si5351a frequency synthesizer.


   The system starts executing the code in a function called setup() and then it
   repeatedly keeps calling loop() forever. All the initialization code is kept in setup()
   and code to continuously sense the tuning knob, the function button, transmit/receive,
   etc is all in the loop() function. If you wish to study the code top down, then scroll
   to the bottom of this file and read your way up.


    The main chip which generates up to three oscillators of various frequencies in the
    Raduino is the Si5351a. To learn more about Si5351a you can download the datasheet
    from www.silabs.com although, strictly speaking it is not a requirement to understand this code.
    Instead, you can look up the Si5351 library written by xxx, yyy. You can download and
    install it from www.url.com to compile this file.
    The Wire.h library is used to talk to the Si5351 and we also declare an instance of
    Si5351 object to control the clocks.
*/
char Version_Text[21] = "   I2C4LT_V6.02b";
char Version_Date[21] = " 08/05/2018 16:50";

/*
    1. Renamed all files with .cpp extension except ubitx20.ino
    2. Creates ubitx20.h to hold constants that are used by multiple files.
    3. Created list of external routines and variable not defined in current file.

    This will be the base for further development.

    Ron - W2CTX
    Jim - W0EB (editor & beta tester)
*/


#include <Arduino.h>

//#include <Wire.h>
#include <i2c_t3.h>
#include "USBHost_t36.h"

#include <EEPROM.h>
#include <TimeLib.h>
#include <SPI.h>
#include <SD.h>
#include "ubitx.h"



extern void si5351bx_setfreq(uint8_t clknum, uint32_t fout);
extern void updateDisplay(void);
extern int  btnDown(void);
extern void doMenu(void);
extern int  enc_read(void);
extern void printLineXI(char linenmbr, char index, char *c);
extern void printLine(char linenmbr, char *c);
extern void printLine1(char *c);
extern void printLine2(char *c);
extern void printLine3(char *c);
extern void printLine4(char *c);
extern void initOscillators(void);
extern void cwKeyer(void);
extern void menuSplitToggle( void );
extern void menuVfoToggle(void );
extern void Connect_Interrupts(void);
extern void Init_Cat( long baud );
extern void Process_CAT_Command( void );
extern void Do_Configuration(void);
extern void DoBaud(void);
extern void menuPaddleSwapToggle( void );
extern void SetIambicA( void );
extern void SetIambicB( void );
extern void SetPaddleSwapN( void);
extern void SetPaddleSwapR( void);
extern void Place_The_Cursor( void);
extern void Display_Voltage( void );
extern void Display_S_Meter();
extern bool Choose_Down(void);
extern bool Enter_Exit_Down(void);
extern bool Put_SDEEPROM( int index, long Value );
extern long Get_SDEEPROM( int index );
extern void SD_Setup();
extern char ubitx_mode_text[6][5];
extern void Blank_Screen( void );
extern void Display_Mem_Selection( void );
extern int Menu_Select_Encoder_Read(void);
extern void Display_S_Meter( void );
extern void Process_Keyboard( void );
extern void Process_CW( void );

void startTx(byte txMode);
void stopTx(void);

extern int Last_Mode;
extern long ubitx_CAT_Long_Baud_rates[8];
extern int ubitx_baud;
extern int Last_Baud;
extern unsigned int SI5351BX_XTALPF;              // 1:6pf  2:8pf  3:10pf
extern bool Cat_Lock;
extern int Memory_Table_Select;
extern unsigned long Working_Copy_Frequencies[10];
extern int   Working_Copy_Modes[10];
extern unsigned long TableFREQS[4][10];
extern int TableMODES[4][10];

extern bool CW_Filter_Present;
extern bool CW_Calibrated;
extern unsigned long CW_Filter_Offset;
extern unsigned long Standard_Offset;

bool Clock_Active = false;
bool Smeter_Active = false;
bool Volts_Active = false;
int Volts_Delay_Ctr = 0;
bool Freq_Incr_Direction = false;
bool Tune_Lock = false;

extern volatile int SmeterReadValue, LastSmeter, SMeter_Value, smeterInterval;


/**
    The main chip which generates up to three oscillators of various frequencies in the
    Raduino is the Si5351a. To learn more about Si5351a you can download the datasheet
    from www.silabs.com although, strictly speaking it is not a requirement to understand this code.

    We no longer use the standard SI5351 library because of its huge overhead due to many unused
    features consuming a lot of program space. Instead of depending on an external library we now use
    Jerry Gaffke's, KE7ER, lightweight standalone minimalist "si5351bx" routines (see further down the
    code). Here are some defines and declarations used by Jerry's routines:
*/




/** Get the LCD I2C Library here:
  https://bitbucket.org/fmalpartida/new-liquidcrystal/downloads
  Move all *_I2C files into a new folder named LiquidCrystal_I2C
  in the Arduino library folder, or use the "install zipped library" feature
*/

//now invoke the LCD_I2C library                   (n5ib: 15 Jan 2018)

//#include <LiquidCrystal_I2C.h>
#include <LiquidCrystal_Wire2_I2C.h>


//specify the pin connections from the LCD to the PCF8574 chip

//////////////////////////////// user defines their address of theit I2C display
////LiquidCrystal_I2C lcd(0x3F, 2, 1, 0, 4, 5, 6, 7, 3, POSITIVE);  //
//LiquidCrystal_I2C lcd(0x2F, 2, 1, 0, 4, 5, 6, 7, 3, POSITIVE);  //

//LiquidCrystal_I2C lcd(0x27, 2, 1, 0, 4, 5, 6, 7, 3, POSITIVE);  //
LiquidCrystal_Wire2_I2C lcd(0x27, 2, 1, 0, 4, 5, 6, 7, 3, POSITIVE);    //


/** arguments for use with PCF8574A I2C I/O Expander chip on LCD I2C interface module:
  I2C 7-bit slave address in hexadecimal, the least significant three bits are user configurable on the interface module
  LCD EN pin to PCF8574A I/O pin P2
  LCD RW pin to PCF8574A I/O pin P1
  LCD RST pin to PCF8574A I/O pin P0
  LCD D4 pin to PCF8574A I/O pin P4
  LCD D5 pin to PCF8574A I/O pin P5
  LCD D6 pin to PCF8574A I/O pin P6
  LCD D7 pin to PCF8574A I/O pin P7
  LCD Backlight pin to PCF8574A I/O pin P3  (optional parameter)
  LCD Backlight control is POSITIVE logic, i.e., "1" turns ON  (optional parameter)
*/


//////////////////////////////////////////////
extern bool Key_Displayed;
USBHost uBITX_usb;
KeyboardController uBITX_Keyboard(uBITX_usb);

extern int Keyboard_Mode;
extern int KeyIndex;
char Keyboard_Input_Buffer[41];


unsigned long Ubitx_Voltage = 0;
unsigned long Current_Ubitx_Voltage = 0;
unsigned long Converted_Ubitx_Voltage = 0;

char Voltage_Text[5];
unsigned long Volt_Divider;

volatile int ubitx_mode;

extern int wpm;


//////////////////// keyer
unsigned volatile char keyerControl = 0;
unsigned volatile char keyerState = 0;
unsigned volatile char IAMBIC = 0x10;  // 0 for Iambic A, 1 for Iambic B
unsigned volatile char PDLSWAP =   0x00;     // 0x00 for normal, 0x08 for swap

///////////////////

char c[21], b[21];
char printBuff[4][21];  //mirrors what is showing on the two lines of the display
char ritOn = 0;
char splitOn = 0;
char vfoActive = VFO_A;
unsigned long vfoA = 07150000L, vfoB = 14200000L, sideTone = 000, usbCarrier;
unsigned long Last_New_Frequency = 0;
int Last_Frequency_Save_Ctr = 0;
unsigned long frequency, ritRxFrequency, ritTxFrequency;  //frequency is the current frequency on the dial
long ritOffset = 0;
long LastritOffset = 0;
long ritOffsetMax = 5000;
unsigned long Freq_Incr[6] = {10, 100, 1000, 10000, 100000, 1000000};
int Freq_Index = 0;

int Freq_Underline[6] = {19, 18, 16, 15, 14, 12};
int Menu_Index = 0;  // new menu

int cwSpeed; //this is actual the dot period in milliseconds
extern int32_t calibration;

/**
   Raduino needs to keep track of current state of the transceiver. These are a few variables that do it
*/
volatile bool inTx = false;                //it is set to 1 if in transmit mode (whatever the reason : CW, PTT or CAT)

byte menuOn = 0;              //set to 1 when the menu is being displayed, if a menu item sets it to zero, the menu is exited
unsigned char txFilter = 0;   //which of the four transmit filters are in use
boolean modeCalibrate = false;//this mode of menus shows extended menus to calibrate the oscillators and choose the proper




/**
   Below are the basic functions that control the uBitx. Understanding the functions before
   you start hacking around
*/
/**
   Select the properly TX harmonic filters
   The four harmonic filters use only three relays
   the four LPFs cover 30-21 MHz, 18 - 14 MHz, 7-10 MHz and 3.5 to 5 MHz
   Briefly, it works like this,
   - When KT1 is OFF, the 'off' position routes the PA output through the 30 MHz LPF
   - When KT1 is ON, it routes the PA output to KT2. Which is why you will see that
     the KT1 is on for the three other cases.
   - When the KT1 is ON and KT2 is off, the off position of KT2 routes the PA output
     to 18 MHz LPF (That also works for 14 MHz)
   - When KT1 is On, KT2 is On, it routes the PA output to KT3
   - KT3, when switched on selects the 7-10 MHz filter
   - KT3 when switched off selects the 3.5-5 MHz filter
   See the circuit to understand this
*/

void setTXFilters(unsigned long freq) {

  if (freq > 21000000L) { // the default filter is with 35 MHz cut-off
    digitalWrite(TX_LPF_A, 0);
    digitalWrite(TX_LPF_B, 0);
    digitalWrite(TX_LPF_C, 0);
  }
  else if (freq >= 14000000L) { //throw the KT1 relay on, the 30 MHz LPF is bypassed and the 14-18 MHz LPF is allowed to go through
    digitalWrite(TX_LPF_A, 1);
    digitalWrite(TX_LPF_B, 0);
    digitalWrite(TX_LPF_C, 0);
  }
  else if (freq > 7000000L) {
    digitalWrite(TX_LPF_A, 1);
    digitalWrite(TX_LPF_B, 1);
    digitalWrite(TX_LPF_C, 0);
  }
  else {
    digitalWrite(TX_LPF_A, 1);
    digitalWrite(TX_LPF_B, 1);
    digitalWrite(TX_LPF_C, 1);
  }
}

/**
   This is the most frequently called function that configures the
   radio to a particular frequency, sideband and sets up the transmit filters

   The transmit filter relays are powered up only during the tx so they dont
   draw any current during RX.

   The carrier oscillator of the detector/modulator is permanently fixed at
   upper sideband. The sideband selection is done by placing the second oscillator
   either 12 MHz below or above the 45 MHz signal thereby inverting the sidebands
   through mixing of the second local oscillator.
*/

void setFrequency(unsigned long xf) {
  unsigned long f;

  f = xf;
  frequency = xf;

  // CW filter if present should be used unless we are transmitting
  usbCarrier = Standard_Offset;  // assume not using CW filter
  if ( !inTx ) {
    if (CW_Filter_Present) {
      if ( (ubitx_mode == MODE_CW) || (ubitx_mode == MODE_CWR))usbCarrier = CW_Filter_Offset;
    }
  }
  if ( ubitx_mode == MODE_CW) f = f - sideTone;
  if ( ubitx_mode == MODE_CWR) f = f + sideTone;

  setTXFilters(f);
  si5351bx_setfreq(0, usbCarrier);

  if ( (ubitx_mode == MODE_SWU) || (ubitx_mode == MODE_USB) || (ubitx_mode == MODE_CW)) {
    si5351bx_setfreq(2, SECOND_OSC_USB - usbCarrier + f + ritOffset);
    si5351bx_setfreq(1, SECOND_OSC_USB);
  } else {
    si5351bx_setfreq(2, SECOND_OSC_LSB + usbCarrier + f + ritOffset);
    si5351bx_setfreq(1, SECOND_OSC_LSB);
  }
  usbCarrier = Standard_Offset;  // assume not using CW filter
}

/**
   startTx is called by the PTT, cw keyer and CAT protocol to
   put the uBitx in TX mode. It takes care of RIT settings, sideband settings
   Note: In cw mode, doesn't key the radio, only puts it in TX mode
*/

void startTx(byte txMode) {
  unsigned long tx_freq;

  if ( (ubitx_mode == MODE_SWU) || (ubitx_mode == MODE_SWL)) {
    printLine2((char *)"No TX in SW Modes");
    delay(500);
  } else if( ubitx_mode != MODE_CWP) {
    if ( splitOn) tx_freq = vfoB;
    else tx_freq = frequency;

    digitalWrite(TX_RX, 1);
    inTx = true;

#ifdef XMIT_LED
    //if ( MODE_USB || MODE_LSB )digitalWrite(ON_AIR, 1); // light up the LED on NANO's pin 13
    digitalWrite(ON_AIR, 1); // light up the LED on NANO's pin 13
#endif


    if (txMode == TX_CW) {
      //turn off the second local oscillator and the BFO
      si5351bx_setfreq(0, 0);
      si5351bx_setfreq(1, 0);
      si5351bx_setfreq(2, tx_freq );
    }
  }
}
void stopTx(void) {
  inTx = false;

  digitalWrite(TX_RX, 0);           //turn off the TX
  si5351bx_setfreq(0, usbCarrier);  //set back the carrier oscillator anyway, CW TX switches it off

#ifdef XMIT_LED
  //if ( MODE_USB || MODE_LSB )digitalWrite(ON_AIR, 0); // extinguish the LED on NANO's pin 13
  digitalWrite(ON_AIR, 0); // extinguish the LED on NANO's pin 13
#endif

  setFrequency(frequency);
}

/**
   ritEnable is called with a frequency parameter that determines
   what the RX frequency will be
*/
void ritEnable(void) {
  if (!ritOn) {
    ritOffset = LastritOffset;
    ritOn = 1;
    setFrequency(frequency);
    updateDisplay();
    Display_Mem_Selection();
    printLineXI(2, 10, (char *)"RIT");
  }
}

// this is called by the RIT menu routine
void ritDisable(void) {
  if (ritOn) {
    ritOn = 0;
    LastritOffset = ritOffset;
    ritOffset = 0;
    printLineXI(2, 10, (char *)"   ");
    setFrequency(frequency);
    updateDisplay();
    Display_Mem_Selection();
  }
}


/**
   Basic User Interface Routines. These check the front panel for any activity
*/


void Save_Radio_State( void ) {

  if (vfoActive == VFO_A) {
    Put_SDEEPROM( VFO_A, frequency );
  } else {
    Put_SDEEPROM( VFO_B, frequency );
  }

}

bool Get_Button( int button ) {

  if (digitalRead(button) == HIGH) return false;
  else return true;

}

// test button for short or long push
int Test_Button( int button, char * title1, char * title2, char * title3, char * title4 ) {
  int i;

  //only if the button is pressed
  if (!Get_Button(button)) return 0;
  delay(100);
  if (!Get_Button(button)) return 0; //debounce
  // OK button is definitely pushed so now detect how long it is pressed.

  if ( Cat_Lock == true ) {
    while ( Get_Button(button)) delay(10);
    return -1;
  } else {
    i = 0;
    while ( Get_Button(button) ) {
      delay(10);
      i++;
      if ( (i > BUTTON_DELAY_1) && (i < BUTTON_DELAY_2) )printLine2(title1);
      if ( (i > BUTTON_DELAY_2) && (i < BUTTON_DELAY_3) )printLine2(title2);
      if ( (i > BUTTON_DELAY_3) && (i < BUTTON_DELAY_4) )printLine2(title3);
      if ( i > BUTTON_DELAY_4) printLine2(title4);
    }
    if ((i > BUTTON_DELAY_1) && (i < BUTTON_DELAY_2) )return 1;
    if ((i > BUTTON_DELAY_2 ) && (i < BUTTON_DELAY_3) )return 2;
    if ((i > BUTTON_DELAY_3 ) && (i < BUTTON_DELAY_4) )return 3;
    if ( i > BUTTON_DELAY_4) return 4;
    else return -1;
  }
}




void checkButton( int button ) {
  int Do_Something;

  if ( Enter_Exit_Down() ) {
    if ( CW_Filter_Present )Do_Something = Test_Button(button, (char*)"MENU", (char*)"CONFIGURATION", (char*)"CW CAL", (char*)"EXIT");
    else Do_Something = Test_Button(button, (char*)"MENU", (char*)"CONFIGURATION", (char*)"EXIT", (char*)"EXIT");
    if (Do_Something != 0 ) {
      if ( Do_Something == 1) {
        Blank_Screen();
        doMenu();
        if (ritOn)    printLineXI(2, 10, (char *)"RIT");
      } else if ( Do_Something == 2 ) {
        Blank_Screen();
        Do_Configuration();
        if (ritOn)    printLineXI(2, 10, (char *)"RIT");
      } else if ( Do_Something == 3 ) {
        if (CW_Filter_Present) Do_CW_CAL();
        printLine2((char *)" ");
      } else if ( Do_Something == 4 ) {
        printLine2((char *)" ");
      }
      Display_Mem_Selection();
    }
  }
}



void doTuning() {
  int s;
  long rx_freq;

  if ( !Tune_Lock ) {
    s = enc_read();
    if (s) {
      if ( vfoActive == VFO_B ) rx_freq = vfoB;
      else rx_freq = vfoA;

      if (s > 4)
        rx_freq += Freq_Incr[Freq_Index];
      else if (s >  2)
        rx_freq += Freq_Incr[Freq_Index];
      else if (s >  0)
        rx_freq +=  Freq_Incr[Freq_Index];
      else if (s <  0)
        rx_freq -= Freq_Incr[Freq_Index];
      else if (s < -2)
        rx_freq -= Freq_Incr[Freq_Index];
      else if (s < -4)
        rx_freq -= Freq_Incr[Freq_Index];

      if ( rx_freq <  400000l) rx_freq =  400000l;
      if ( rx_freq > 30000000l) rx_freq = 30000000l;

      if ( vfoActive == VFO_B )vfoB = rx_freq;
      else vfoA = rx_freq;
      frequency = rx_freq;
      setFrequency(frequency);
      updateDisplay();
    }
  }
}
void Do_CW_CAL( void ) {
  int s, i;
  int Local_Freq_Index = 2;


  printLine1((char*)"Choose - Exit");
  printLine2((char*)"tune until ");
  printLine3((char*)"sounds good ");
  printLine4((char*)"  ");
  //CW_Filter_Offset = Get_SDEEPROM( CW_FILTER_OFFSET );
  CW_Filter_Offset = 8000000l;
  si5351bx_setfreq(0, CW_Filter_Offset);
  setFrequency(frequency);
  while (!Choose_Down()) {
    s = enc_read();
    if (s) {
      if (s > 4)
        CW_Filter_Offset += Freq_Incr[Local_Freq_Index];
      else if (s >  2)
        CW_Filter_Offset += Freq_Incr[Local_Freq_Index];
      else if (s >  0)
        CW_Filter_Offset +=  Freq_Incr[Local_Freq_Index];
      else if (s <  0)
        CW_Filter_Offset -= Freq_Incr[Local_Freq_Index];
      else if (s < -2)
        CW_Filter_Offset -= Freq_Incr[Local_Freq_Index];
      else if (s < -4)
        CW_Filter_Offset -= Freq_Incr[Local_Freq_Index];

      si5351bx_setfreq(0, CW_Filter_Offset);
      setFrequency(frequency);
    }
    // check freq incr
    if (FreqIncr_RIT_btnDown()) {
      i = 0;
      while ( FreqIncr_RIT_btnDown() ) {
        delay(10);
        i++;
      }
      if ( i > 1) {
        Local_Freq_Index = Local_Freq_Index - 1;
        if ( Local_Freq_Index < 0 )Local_Freq_Index = 5;
      }
    }
    memset(c, 0, sizeof(c));
    memset(b, 0, sizeof(b));
    ultoa(CW_Filter_Offset, b, DEC);
    if (strlen(b) < 7) {
      strcat(&c[0], "00.");
      strncat(c, &b[0], 3);
      strcat(c, ".");
      strncat(c, &b[3], 2);
    } else if (strlen(b) < 8) {
      strcat(&c[0], "0");
      strncat(c, &b[0], 1);
      strcat(c, ".");
      strncat(c, &b[1], 3);
      strcat(c, ".");
      strncat(c, &b[4], 2);
    } else {
      strncat(c, b, 2);
      strcat(c, ".");
      strncat(c, &b[2], 3);
      strcat(c, ".");
      strncat(c, &b[5], 2);
    }

    printLineXI(3, 11, c);
    lcd.setCursor(Freq_Underline[Local_Freq_Index], 3);
  }
  while (Choose_Down())delay(10);

  Put_SDEEPROM( CW_FILTER_OFFSET, CW_Filter_Offset );
  Put_SDEEPROM( CW_FILTER_CALIBRATED, true );
  CW_Calibrated = true;

  printLine3((char*)"  ");
  printLine1((char*)"  ");
  updateDisplay();

}


/**
   RIT only steps back and forth by 100 Hz at a time
*/
void doRIT() {

  int knob = enc_read();

  if (knob) {
    if (knob < 0) {
      ritOffset -= 100l;
      if ( abs(ritOffset) > ritOffsetMax) ritOffset = -ritOffsetMax;
    } else if (knob > 0) {
      ritOffset += 100;
      if ( ritOffset > ritOffsetMax) ritOffset = ritOffsetMax;
    }
    setFrequency(frequency);
    updateDisplay();
  }
}


/**
   The settings are read from EEPROM. The first time around, the values may not be
   present or out of range, in this case, some intelligent defaults are copied into the
   variables.
*/
void initSettings() {
  int i;

  //read the settings from the eeprom and restore them

  calibration = Get_SDEEPROM( MASTER_CAL );
  usbCarrier  = Get_SDEEPROM( USB_CAL );

  Standard_Offset = usbCarrier;


  SI5351BX_XTALPF = Get_SDEEPROM( XTALPF );
  if ( (SI5351BX_XTALPF < 1) || (SI5351BX_XTALPF > 3)) {
    SI5351BX_XTALPF = 2;
    Put_SDEEPROM( XTALPF, SI5351BX_XTALPF );
  }

  vfoA = Get_SDEEPROM( VFO_A );
  if ( vfoA <   400000l ) vfoA = 400000l;
  if ( vfoA > 30000000l ) vfoA = 30000000l;

  vfoB = Get_SDEEPROM( VFO_B );
  if ( vfoB <   400000l ) vfoB = 400000l;
  if ( vfoB > 30000000l ) vfoB = 30000000l;

  sideTone = Get_SDEEPROM( CW_SIDETONE );
  if ( sideTone < 100 || sideTone > 2000 ) sideTone = 800;

  wpm = Get_SDEEPROM( IAMBIC_MODE );
  IAMBIC = (volatile unsigned char)wpm;
  if ( IAMBIC == 0 )SetIambicA();
  else SetIambicB();

  wpm = Get_SDEEPROM( PDL_SWAP_VAL );
  PDLSWAP = (volatile unsigned char)wpm;
  if ( PDLSWAP == 1 ) SetPaddleSwapR();
  else SetPaddleSwapN();

  wpm = Get_SDEEPROM( MODE );
  if ( wpm < 0 ) wpm = 0;
  if ( wpm > 6 ) wpm = 6;
  ubitx_mode = wpm;
  Last_Mode = wpm;

  wpm = Get_SDEEPROM( VOLTS );
  if ( wpm == 1 ) {
    Volts_Active = true;
  } else {
    Volts_Active = false;
  }

  ubitx_baud = Get_SDEEPROM( BAUD );
  if ( ubitx_baud < 0 ) ubitx_baud = 5;
  if ( ubitx_baud > 7 ) ubitx_baud = 5;
  Last_Baud = ubitx_baud;

  wpm = Get_SDEEPROM( CW_SPEED );
  if ( wpm < 5 ) wpm = 5;
  if ( wpm > 50 ) wpm = 50;
  cwSpeed = 600 / wpm;

  Freq_Index = Get_SDEEPROM( FREQ_INCR );
  if ( (Freq_Index < 0) || (Freq_Index > 5)) {
    Freq_Index = 2;
    Put_SDEEPROM( FREQ_INCR, Freq_Index );
    delay(200);
  }



  Smeter_Active = Get_SDEEPROM( SMETER_ACTIVE );
  Clock_Active = Get_SDEEPROM( CLOCK_ACTIVE );


  Volt_Divider = Get_SDEEPROM( VOLT_DIVIDE );
  if ( (Volt_Divider > 2000) || (Volt_Divider <= 0 )) Volt_Divider = 982;


  Memory_Table_Select = Get_SDEEPROM( SELECTED_MEM_TABLE );
  if ( (Memory_Table_Select > 3) || (Memory_Table_Select < 0))Memory_Table_Select = 0;
  for (i = 0; i < 10; i++) {
    Working_Copy_Frequencies[i] = Get_SDEEPROM( ((Memory_Table_Select + 1) * 100 ) + (i * 4) );
    Working_Copy_Modes[i] = Get_SDEEPROM( ((Memory_Table_Select + 1) * 100 ) + 40 + (i * 4) );
  }

  CW_Filter_Present = Get_SDEEPROM( CW_FILTER_PRESENT );
  CW_Calibrated = Get_SDEEPROM( CW_FILTER_CALIBRATED );
  CW_Filter_Offset = Get_SDEEPROM( CW_FILTER_OFFSET );
  if ( !CW_Filter_Present )CW_Filter_Offset = usbCarrier;

}




void initPorts() {


  pinMode(ENC_A, INPUT_PULLUP);
  pinMode(ENC_B, INPUT_PULLUP);
  pinMode(FBUTTON, INPUT_PULLUP);
  pinMode(AB_SPLIT, INPUT_PULLUP);
  pinMode(UBIT_VCC, INPUT );
  pinMode(SMETERREADPIN, INPUT );

  // init CW digital pins
  pinMode(DIGITAL_DOT, INPUT_PULLUP);
  pinMode(DIGITAL_DASH, INPUT_PULLUP);
  pinMode(DIGITAL_KEY, INPUT_PULLUP);

  //Menu Enter/Select Push button D28
  pinMode(MENU_ENTER_BUTTON, INPUT_PULLUP);

  //Menu Choose  Push button D29
  pinMode(MENU_CHOOSE_BUTTON, INPUT_PULLUP);
  pinMode(MENU_SEL_ENC_A, INPUT_PULLUP);
  pinMode(MENU_SEL_ENC_B, INPUT_PULLUP);

  pinMode(MEMORY_TABLES_BUTTON, INPUT_PULLUP);
  pinMode(MEMORY_CHANNEL_BUTTON, INPUT_PULLUP);
  pinMode(DISPLAY_SEQUENCE_BUTTON, INPUT_PULLUP);

  pinMode(CW_TONE, OUTPUT);
  digitalWrite(CW_TONE, 0);

  pinMode(TX_RX, OUTPUT);
  digitalWrite(TX_RX, 0);

  pinMode(TX_LPF_A, OUTPUT);
  pinMode(TX_LPF_B, OUTPUT);
  pinMode(TX_LPF_C, OUTPUT);
  digitalWrite(TX_LPF_A, 0);
  digitalWrite(TX_LPF_B, 0);
  digitalWrite(TX_LPF_C, 0);

  pinMode(CW_KEY, OUTPUT);
  digitalWrite(CW_KEY, 0);


#ifdef XMIT_LED
  pinMode(LED_BUILTIN, OUTPUT); //use the LED on D13 as an "On the Air" indicator
  digitalWrite(LED_BUILTIN, 0);  // turn it off to begin with
#endif

}


int Last_Second = 0;
int Last_Minute = 0;

void digitalClockDisplay() {
  char Time_Buffer[21];
  uint8_t oldSREG;



  // digital clock display of the time
  if ( Last_Second != second()) {
    Last_Second = second();
    if ( Clock_Active ) {
      if (  Keyboard_Mode != KEY_CW ) {
        strcpy(Time_Buffer, "");
        sprintf(Time_Buffer, "%02d:%02d:%02d", hour(), minute(), second());
        printLineXI(0, 12, Time_Buffer);
      }
    }
    strcpy(Time_Buffer, "");
    // process VFO save
    Last_Frequency_Save_Ctr--;
    if ( Last_Frequency_Save_Ctr <= 0 ) {
      Last_Frequency_Save_Ctr = 5;
      if ( Last_New_Frequency != frequency ) {
        if ( vfoActive == VFO_B ) {
          Put_SDEEPROM( VFO_B, frequency );
        } else {
          Put_SDEEPROM( VFO_A, frequency );
        }
        Last_New_Frequency = frequency;
      }
    }

    if ( Volts_Delay_Ctr <= 0) {
      Volts_Delay_Ctr = 5;
      if ( Volts_Active ) {


        oldSREG = SREG;
        sei();
        Ubitx_Voltage = analogRead(UBIT_VCC);
        Volts_Active = true;
        Ubitx_Voltage = analogRead(UBIT_VCC);
        SREG = oldSREG;


        if ( Volts_Active != Current_Ubitx_Voltage ) {
          Current_Ubitx_Voltage = Volts_Active;
          Display_Voltage();
        }
      }
    } else {
      Volts_Delay_Ctr--;
    }
    Place_The_Cursor();
  }
}



time_t getTeensy3Time()
{
  return Teensy3Clock.get();
}

void uBITX_Keyboard_Press(int key) {

  Keyboard_Input_Buffer[0] = uBITX_Keyboard.getKey();
  KeyIndex = 1;

}




//////////////////////////////////////////////////////////////////////////////////////////
void setup() {

  Wire.begin(I2C_MASTER, 0x00, I2C_PINS_3_4, I2C_PULLUP_EXT, 400000);

  lcd.begin(20, 4);
  lcd.clear();
  //we print this line so this shows up even if the Raduino
  //crashes later in the code
  printLine1(Version_Text);
  printLine2((char*)"  W0EB  and  W2CTX");
  printLine3(Version_Date);
  delay(2000);
  lcd.clear();


  SD_Setup();

  initSettings();
  initPorts();
  initOscillators();

  frequency = vfoA;
  setFrequency(vfoA);


  keyerState = IDLE;
  keyerControl = IAMBIC; // 0x10 for B Or 0 for A

  lcd.cursor();

  uBITX_usb.begin();
  uBITX_Keyboard.attachPress(uBITX_Keyboard_Press);
  uBITX_usb.Task();

  // set the Time library to use Teensy 3.0's RTC to keep time
  setSyncProvider(getTeensy3Time);
  Teensy3Clock.get();
  printLine2((char*)" ");
  digitalClockDisplay();

  Display_Mem_Selection();

  Init_Cat( ubitx_CAT_Long_Baud_rates[ubitx_baud] );

  Connect_Interrupts();

  updateDisplay();

}



bool FreqIncr_RIT_btnDown( void ) {

  if (digitalRead(FBUTTON) == HIGH) return false;
  else return true;

}



void checkFreqIncr_RIT( void ) {
  int i, j;

  if (FreqIncr_RIT_btnDown()) {
    delay(100);
    if (!FreqIncr_RIT_btnDown()) return; //debounce
    i = 0;
    while ( FreqIncr_RIT_btnDown() ) {
      delay(10);
      i++;
      if ( i > BUTTON_DELAY_4 ) {
        printLineXI(2, 0, (char*)"EXIT   ");
        lcd.setCursor(0, 1);
      } else if ( i > BUTTON_DELAY_3 ) {
        if (Tune_Lock)printLineXI(2, 0, (char*)"LOCKOFF");
        else printLineXI(2, 0, (char*)"LOCKON ");
        lcd.setCursor(0, 1);
      } else if ( i > BUTTON_DELAY_2 ) {
        if (ritOn)printLineXI(2, 0, (char*)"RIT OFF");
        else printLineXI(2, 0, (char*)"RIT ON ");
        lcd.setCursor(0, 1);
      } else if ( i > BUTTON_DELAY_1 ) {
        j = Freq_Index;
        Freq_Index = Freq_Index + 1;
        if ( Freq_Index > 5 )Freq_Index = 0;
        if (vfoActive == VFO_A)lcd.setCursor(Freq_Underline[Freq_Index], 3);
        else lcd.setCursor(Freq_Underline[Freq_Index], 2);
        Freq_Index = j;
      }
    }
    if ( i > BUTTON_DELAY_4 ) {
      printLineXI(2,0,(char*)"    ");
    } else if ( i > BUTTON_DELAY_3 ) {
      delay(10);
      while ( FreqIncr_RIT_btnDown() ) delay(1);
      if ( Tune_Lock ) {
        printLineXI(2, 0, (char*)"       ");
        Tune_Lock = false;
      } else {
        if (ritOn) {
          ritDisable();
          printLineXI(2, 0, (char *)"        ");

        }
        Tune_Lock = true;
      }
    } else if ( i > BUTTON_DELAY_2 ) {
      if ( (splitOn == 0) && (!Tune_Lock)) {
        if (ritOn) ritDisable();
        else ritEnable();
        printLineXI(2, 0, (char *)"        ");
      } else if ( !Tune_Lock )printLineXI(2, 0, (char*)"       ");
      else printLineXI(2, 0, (char*)"LOCKON ");
      updateDisplay();
    } else if ( i > BUTTON_DELAY_1 ) {
      if ( !ritOn) {
        delay(100);
        Freq_Index = Freq_Index + 1;
        if ( Freq_Index > 5 )Freq_Index = 0;
        Put_SDEEPROM( FREQ_INCR, Freq_Index );
        delay(200);
        updateDisplay();
      }
    } else {
      if ( !ritOn) {
        delay(100);
        Freq_Index = Freq_Index - 1;
        if ( Freq_Index < 0 )Freq_Index = 5;
        Put_SDEEPROM( FREQ_INCR, Freq_Index );
        delay(200);
        updateDisplay();
      }
    }
  }
  //printLine3((char*)" ");

}




bool AB_SPLIT_btnDown( void ) {

  if (digitalRead(AB_SPLIT) == HIGH) return false;
  else return true;

}


void checkAB_SPLIT( void ) {
  int i;

  if (AB_SPLIT_btnDown()) {
    //only if the button is pressed
    if (!AB_SPLIT_btnDown()) return;
    delay(100);
    if (!AB_SPLIT_btnDown()) return; //debounce
    // OK button is definitely pushed so now detect how long it is pressed.
    i = 0;
    while ( AB_SPLIT_btnDown() ) {
      delay(10);
      i++;
      if ( i > BUTTON_DELAY_2 ) {
        if (splitOn)printLine2((char*)"SPLIT OFF");
        else printLine2((char*)"SPLIT ON");
        lcd.setCursor(0, 1);
      }
    }
    if ( i > BUTTON_DELAY_2 ) {
      if ( !ritOn) menuSplitToggle();
    }
    else menuVfoToggle();
    if ( Tune_Lock )printLineXI(2, 0, (char*)"LOCKON ");
    else printLine2((char*)" ");

    updateDisplay();
    Display_Mem_Selection();
  }
}



bool MEMORY_CHANNEL_BUTTON_Down(void) {

  if (digitalRead(MEMORY_CHANNEL_BUTTON) == HIGH) return false;
  else return true;

}


extern int Mem_Number;
extern bool Choose_Down(void);

void Process_Memory_Select( void) {
  int Ctr;
  int Mem_Table_Ptr;
  int i;


  if (MEMORY_CHANNEL_BUTTON_Down()) {
    Ctr = 0;
    while (MEMORY_CHANNEL_BUTTON_Down()) {
      Ctr = Ctr + 1;
      if ( Ctr > 260 )printLineXI(0, 0, (char*)"Exit      ");
      else if ( Ctr > 160 )printLineXI(0, 0, (char*)"Store     ");
      else if ( Ctr > 60 )printLineXI(0, 0, (char*)"Load      ");
      delay(10);
    }
    if ( Ctr < 60 ) {
      Mem_Number = Mem_Number + 1;
      if ( Mem_Number > 9 ) Mem_Number = 0;
      printLineXI(1, 0, (char*)"         ");
    } else if ( Ctr > 260 ) {
      // exit do nothing
    } else if ( Ctr > 160 ) {
      if (vfoActive == VFO_B)Working_Copy_Frequencies[Mem_Number] = vfoB;
      else Working_Copy_Frequencies[Mem_Number] = vfoA;
      Working_Copy_Modes[Mem_Number] = ubitx_mode;
      Mem_Table_Ptr = ( Memory_Table_Select + 1 ) * 100;
      for (i = 0; i < 10; i++) {
        Put_SDEEPROM( (Mem_Table_Ptr + (i * 4)), Working_Copy_Frequencies[i] );
        Put_SDEEPROM( (Mem_Table_Ptr + 40 + (i * 4)), Working_Copy_Modes[i] );
      }
    } else if ( Ctr > 60 ) {
      ubitx_mode = Working_Copy_Modes[Mem_Number];
      if (vfoActive == VFO_B)vfoB = Working_Copy_Frequencies[Mem_Number];
      else vfoA = Working_Copy_Frequencies[Mem_Number];
      frequency = Working_Copy_Frequencies[Mem_Number];
      Put_SDEEPROM( MODE, ubitx_mode );
      delay(200);
      setFrequency(frequency);
    }
    Display_Mem_Selection();
    updateDisplay();
  }

}

bool MEMORY_TABLES_BUTTON_Down(void) {

  if (digitalRead(MEMORY_TABLES_BUTTON) == HIGH) return false;
  else return true;

}


extern char Memory_Tables[4][4];

void Process_Memory_Tables( void) {
  int Ctr;
  int Timer_Ctr;
  int Mem_Table;
  int i;
  int Mem_Table_Ptr;


  if (MEMORY_TABLES_BUTTON_Down()) {
    Mem_Table = Memory_Table_Select;
    Ctr = 0;
    Timer_Ctr = 3000;
    while ( Timer_Ctr > 0 ) {
      if ( MEMORY_TABLES_BUTTON_Down()) {
        while (MEMORY_TABLES_BUTTON_Down()) {
          delay(10);
          Ctr = Ctr + 1;
          if ( Ctr > 160 )printLineXI(1, 0, (char*)"Exit      ");
          else if ( Ctr > 60 )printLineXI(1, 0, (char*)"Load      ");
        }
        if ( Ctr < 60 ) {
          Timer_Ctr = 3000;
          Mem_Table = Mem_Table + 1;
          if ( Mem_Table > 3 ) Mem_Table = 0;
          printLineXI(1, 0, (char*)"         ");
          printLineXI(1, 0, Memory_Tables[Mem_Table]);
          Ctr = 0;
        } else if ( Ctr > 160 ) {
          // exit do nothing
          Timer_Ctr = 0;
        } else if ( Ctr > 60 ) {
          Memory_Table_Select = Mem_Table;
          Mem_Table_Ptr = ( Mem_Table + 1 ) * 100;
          for (i = 0; i < 10; i++) {
            Working_Copy_Frequencies[i] = Get_SDEEPROM( Mem_Table_Ptr + (i * 4) );
            Working_Copy_Modes[i] = Get_SDEEPROM( Mem_Table_Ptr + 40 + (i * 4) );
          }
          Put_SDEEPROM( SELECTED_MEM_TABLE, Mem_Table );

          Timer_Ctr = 0;
        }
      } else {
        delay(10);
        Timer_Ctr = Timer_Ctr - 10;
      }
    }
    Display_Mem_Selection();
    updateDisplay();
  }


}

////////////////////////////////////////////////////////   L  O  O  P  /////////////////////////
void loop() {
  uint8_t oldSREG;

  Process_CAT_Command();
  if ( KeyIndex == 1 ) Process_Keyboard();



  //do tasks only when not tranmsitting
  if (!inTx) {
    if ( Smeter_Active == true ) {
      oldSREG = SREG;
      sei();

      if ( smeterInterval <= 0 ) {
        smeterInterval = 150;
        //smeterInterval = 500;
        SREG = oldSREG;
        Display_S_Meter();
        oldSREG = SREG;
      }
      sei();
    }
    digitalClockDisplay()  ;
    if ( Cat_Lock == false ) {
      Process_CAT_Command();
      checkAB_SPLIT();
      Process_CAT_Command();
      checkButton(MENU_ENTER_BUTTON);
      Process_CAT_Command();
      checkFreqIncr_RIT();
      Process_CAT_Command();
      if (ritOn) {
        doRIT();
        Process_CAT_Command();
      } else {
        doTuning();
      }
      Process_Memory_Select();
      Process_Memory_Tables();
    } else {
      doTuning();
      checkButton(FBUTTON);
    }
  }
}


